"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const fs_ = require("fs");
const path_ = require("path");
const ts = require("typescript");
const logging_1 = require("./logging");
const match = require("./match-files");
const util = require("./util");
/**
 * In-memory file system, can be served as a ParseConfigHost (thus allowing listing files that belong to project based on tsconfig.json options)
 */
class InMemoryFileSystem {
    constructor(path, logger = new logging_1.NoopLogger()) {
        this.logger = logger;
        /**
         * Contains a Map of all URIs that exist in the workspace, optionally with a content.
         * File contents for URIs in it do not neccessarily have to be fetched already.
         */
        this.files = new Map();
        this.path = path;
        this.overlay = new Map();
        this.rootNode = { file: false, children: new Map() };
    }
    /**
     * Returns an IterableIterator for all URIs known to exist in the workspace (content loaded or not)
     */
    uris() {
        return this.files.keys();
    }
    /**
     * Adds a file to the local cache
     *
     * @param uri The URI of the file
     * @param content The optional content
     */
    add(uri, content) {
        // Make sure not to override existing content with undefined
        if (content !== undefined || !this.files.has(uri)) {
            this.files.set(uri, content);
        }
        // Add to directory tree
        const filePath = path_.posix.relative(this.path, util.uri2path(uri));
        const components = filePath.split('/');
        let node = this.rootNode;
        for (const [i, component] of components.entries()) {
            const n = node.children.get(component);
            if (!n) {
                if (i < components.length - 1) {
                    const n = { file: false, children: new Map() };
                    node.children.set(component, n);
                    node = n;
                }
                else {
                    node.children.set(component, { file: true, children: new Map() });
                }
            }
            else {
                node = n;
            }
        }
    }
    /**
     * Returns true if the given file is known to exist in the workspace (content loaded or not)
     *
     * @param uri URI to a file
     */
    has(uri) {
        return this.files.has(uri) || this.fileExists(util.uri2path(uri));
    }
    /**
     * Returns the file content for the given URI.
     * Will throw an Error if no available in-memory.
     * Use FileSystemUpdater.ensure() to ensure that the file is available.
     *
     * TODO take overlay into account
     */
    getContent(uri) {
        let content = this.files.get(uri);
        if (content === undefined) {
            content = exports.typeScriptLibraries.get(util.uri2path(uri));
        }
        if (content === undefined) {
            throw new Error(`Content of ${uri} is not available in memory`);
        }
        return content;
    }
    /**
     * Tells if a file denoted by the given name exists in the workspace (does not have to be loaded)
     *
     * @param path File path or URI (both absolute or relative file paths are accepted)
     */
    fileExists(path) {
        return this.readFileIfExists(path) !== undefined || this.files.has(path) || this.files.has(util.path2uri(this.path, path));
    }
    /**
     * @param path file path (both absolute or relative file paths are accepted)
     * @return file's content in the following order (overlay then cache).
     * If there is no such file, returns empty string to match expected signature
     */
    readFile(path) {
        return this.readFileIfExists(path) || '';
    }
    /**
     * @param path file path (both absolute or relative file paths are accepted)
     * @return file's content in the following order (overlay then cache).
     * If there is no such file, returns undefined
     */
    readFileIfExists(path) {
        let content = this.overlay.get(path);
        if (content !== undefined) {
            return content;
        }
        const rel = path_.posix.relative('/', path);
        content = this.overlay.get(rel);
        if (content !== undefined) {
            return content;
        }
        // TODO This assumes that the URI was a file:// URL.
        //      In reality it could be anything, and the first URI matching the path should be used.
        //      With the current Map, the search would be O(n), it would require a tree to get O(log(n))
        content = this.files.get(util.path2uri(this.path, path));
        if (content !== undefined) {
            return content;
        }
        return exports.typeScriptLibraries.get(path);
    }
    /**
     * Invalidates temporary content denoted by the given path
     * @param path path to a file relative to project root
     */
    didClose(path) {
        this.overlay.delete(path);
    }
    /**
     * Adds temporary content denoted by the given path
     * @param path path to a file relative to project root
     */
    didSave(path) {
        const content = this.readFileIfExists(path);
        if (content !== undefined) {
            this.add(util.path2uri('', path), content);
        }
    }
    /**
     * Updates temporary content denoted by the given path
     * @param path path to a file relative to project root
     */
    didChange(path, text) {
        this.overlay.set(path, text);
    }
    /**
     * Called by TS service to scan virtual directory when TS service looks for source files that belong to a project
     */
    readDirectory(rootDir, extensions, excludes, includes) {
        return match.matchFiles(rootDir, extensions, excludes, includes, true, this.path, p => this.getFileSystemEntries(p));
    }
    /**
     * Called by TS service to scan virtual directory when TS service looks for source files that belong to a project
     */
    getFileSystemEntries(path) {
        const ret = { files: [], directories: [] };
        let node = this.rootNode;
        const components = path.split('/').filter(c => c);
        if (components.length !== 1 || components[0]) {
            for (const component of components) {
                const n = node.children.get(component);
                if (!n) {
                    return ret;
                }
                node = n;
            }
        }
        node.children.forEach((value, name) => {
            if (value.file) {
                ret.files.push(name);
            }
            else {
                ret.directories.push(name);
            }
        });
        return ret;
    }
    trace(message) {
        this.logger.log(message);
    }
}
exports.InMemoryFileSystem = InMemoryFileSystem;
/**
 * Iterates over in-memory cache calling given function on each node until callback signals abort or all nodes were traversed
 */
function walkInMemoryFs(fs, rootdir, walkfn) {
    const err = walkfn(rootdir, true);
    if (err) {
        if (err === exports.skipDir) {
            return;
        }
        return err;
    }
    const { files, directories } = fs.getFileSystemEntries(rootdir);
    for (const file of files) {
        const err = walkfn(path_.posix.join(rootdir, file), false);
        if (err) {
            return err;
        }
    }
    for (const dir of directories) {
        const err = walkInMemoryFs(fs, path_.posix.join(rootdir, dir), walkfn);
        if (err) {
            return err;
        }
    }
    return;
}
exports.walkInMemoryFs = walkInMemoryFs;
/**
 * TypeScript library files fetched from the local file system (bundled TS)
 */
exports.typeScriptLibraries = new Map();
/**
 * Fetching TypeScript library files from local file system
 */
const path = path_.dirname(ts.getDefaultLibFilePath({ target: ts.ScriptTarget.ES2015 }));
fs_.readdirSync(path).forEach(file => {
    const fullPath = path_.join(path, file);
    if (fs_.statSync(fullPath).isFile()) {
        exports.typeScriptLibraries.set(util.toUnixPath(fullPath), fs_.readFileSync(fullPath).toString());
    }
});
/**
 * @param path file path
 * @return true if given file belongs to bundled TypeScript libraries
 */
function isTypeScriptLibrary(path) {
    return exports.typeScriptLibraries.has(util.toUnixPath(path));
}
exports.isTypeScriptLibrary = isTypeScriptLibrary;
/**
 * Indicates that tree traversal function should stop
 */
exports.skipDir = {
    name: 'WALK_FN_SKIP_DIR',
    message: ''
};
//# sourceMappingURL=memfs.js.map