"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const rxjs_1 = require("@reactivex/rxjs");
const iterare_1 = require("iterare");
const opentracing_1 = require("opentracing");
const os = require("os");
const path_ = require("path");
const ts = require("typescript");
const url = require("url");
const logging_1 = require("./logging");
const memfs_1 = require("./memfs");
const util = require("./util");
/**
 * ProjectManager translates VFS files to one or many projects denoted by [tj]config.json.
 * It uses either local or remote file system to fetch directory tree and files from and then
 * makes one or more LanguageService objects. By default all LanguageService objects contain no files,
 * they are added on demand - current file for hover or definition, project's files for references and
 * all files from all projects for workspace symbols.
 */
class ProjectManager {
    /**
     * @param rootPath root path as passed to `initialize`
     * @param inMemoryFileSystem File system that keeps structure and contents in memory
     * @param strict indicates if we are working in strict mode (VFS) or with a local file system
     * @param traceModuleResolution allows to enable module resolution tracing (done by TS compiler)
     */
    constructor(rootPath, inMemoryFileSystem, updater, strict, traceModuleResolution, logger = new logging_1.NoopLogger()) {
        this.logger = logger;
        /**
         * A URI Map from file to files referenced by the file, so files only need to be pre-processed once
         */
        this.referencedFiles = new Map();
        this.rootPath = util.toUnixPath(rootPath);
        this.configs = new Map();
        this.updater = updater;
        this.localFs = inMemoryFileSystem;
        this.versions = new Map();
        this.strict = strict;
        this.traceModuleResolution = traceModuleResolution || false;
    }
    /**
     * Disposes the object and cancels any asynchronous operations that are still active
     */
    dispose() {
        // TODO unsubscribe subscriptions
    }
    /**
     * @return root path (as passed to `initialize`)
     */
    getRemoteRoot() {
        return this.rootPath;
    }
    /**
     * @return local side of file content provider which keeps cached copies of fethed files
     */
    getFs() {
        return this.localFs;
    }
    /**
     * @param filePath file path (both absolute or relative file paths are accepted)
     * @return true if there is a fetched file with a given path
     */
    hasFile(filePath) {
        return this.localFs.fileExists(filePath);
    }
    /**
     * @return all sub-projects we have identified for a given workspace.
     * Sub-project is mainly a folder which contains tsconfig.json, jsconfig.json, package.json,
     * or a root folder which serves as a fallback
     */
    configurations() {
        return this.configs.values();
    }
    /**
     * Ensures that the module structure of the project exists in memory.
     * TypeScript/JavaScript module structure is determined by [jt]sconfig.json,
     * filesystem layout, global*.d.ts and package.json files.
     * Then creates new ProjectConfigurations, resets existing and invalidates file references.
     */
    ensureModuleStructure(childOf = new opentracing_1.Span()) {
        if (this.ensuredModuleStructure) {
            return this.ensuredModuleStructure;
        }
        this.ensuredModuleStructure = (() => __awaiter(this, void 0, void 0, function* () {
            const span = childOf.tracer().startSpan('Ensure module structure', { childOf });
            try {
                yield this.updater.ensureStructure();
                // Ensure content of all all global .d.ts, [tj]sconfig.json, package.json files
                yield Promise.all(iterare_1.default(this.localFs.uris())
                    .filter(uri => util.isGlobalTSFile(uri) || util.isConfigFile(uri) || util.isPackageJsonFile(uri))
                    .map(uri => this.updater.ensure(uri)));
                // Scan for [tj]sconfig.json files
                this.createConfigurations();
                // Reset all compilation state
                // TODO utilize incremental compilation instead
                for (const config of this.configs.values()) {
                    config.reset();
                }
                // Require re-processing of file references
                this.invalidateReferencedFiles();
            }
            catch (err) {
                this.ensuredModuleStructure = undefined;
                span.setTag('error', true);
                span.log({ 'event': 'error', 'error.object': err, 'message': err.message, 'stack': err.stack });
                throw err;
            }
            finally {
                span.finish();
            }
        }))();
        return this.ensuredModuleStructure;
    }
    /**
     * Invalidates caches for `ensureModuleStructure`, `ensureAllFiles` and `insureOwnFiles`
     */
    invalidateModuleStructure() {
        this.ensuredModuleStructure = undefined;
        this.ensuredAllFiles = undefined;
        this.ensuredOwnFiles = undefined;
    }
    /**
     * Ensures all files not in node_modules were fetched.
     * This includes all js/ts files, tsconfig files and package.json files.
     * Invalidates project configurations after execution
     */
    ensureOwnFiles(childOf = new opentracing_1.Span()) {
        return __awaiter(this, void 0, void 0, function* () {
            const span = childOf.tracer().startSpan('Ensure own files', { childOf });
            if (!this.ensuredOwnFiles) {
                this.ensuredOwnFiles = (() => __awaiter(this, void 0, void 0, function* () {
                    yield this.updater.ensureStructure(span);
                    yield Promise.all(iterare_1.default(this.localFs.uris())
                        .filter(uri => !uri.includes('/node_modules/') && util.isJSTSFile(uri) || util.isConfigFile(uri) || util.isPackageJsonFile(uri))
                        .map(uri => this.updater.ensure(uri)));
                    this.createConfigurations();
                }))();
            }
            try {
                yield this.ensuredOwnFiles;
            }
            catch (err) {
                this.ensuredOwnFiles = undefined;
                span.setTag('error', true);
                span.log({ 'event': 'error', 'error.object': err, 'message': err.message, 'stack': err.stack });
                throw err;
            }
            finally {
                span.finish();
            }
        });
    }
    /**
     * Ensures all files were fetched from the remote file system.
     * Invalidates project configurations after execution
     */
    ensureAllFiles(childOf = new opentracing_1.Span()) {
        if (this.ensuredAllFiles) {
            return this.ensuredAllFiles;
        }
        this.ensuredAllFiles = (() => __awaiter(this, void 0, void 0, function* () {
            const span = childOf.tracer().startSpan('Ensure all files', { childOf });
            try {
                yield this.updater.ensureStructure(span);
                yield Promise.all(iterare_1.default(this.localFs.uris())
                    .filter(uri => util.isJSTSFile(uri) || util.isConfigFile(uri) || util.isPackageJsonFile(uri))
                    .map(uri => this.updater.ensure(uri)));
                this.createConfigurations();
            }
            catch (err) {
                this.ensuredAllFiles = undefined;
                span.setTag('error', true);
                span.log({ 'event': 'error', 'error.object': err, 'message': err.message, 'stack': err.stack });
                throw err;
            }
            finally {
                span.finish();
            }
        }))();
        return this.ensuredAllFiles;
    }
    /**
     * Recursively collects file(s) dependencies up to given level.
     * Dependencies are extracted by TS compiler from import and reference statements
     *
     * Dependencies include:
     * - all the configuration files
     * - files referenced by the given file
     * - files included by the given file
     *
     * The return values of this method are not cached, but those of the file fetching and file processing are.
     *
     * @param uri File to process
     * @param maxDepth Stop collecting when reached given recursion level
     * @param ignore Tracks visited files to prevent cycles
     * @param childOf OpenTracing parent span for tracing
     * @return Observable of file URIs ensured
     */
    ensureReferencedFiles(uri, maxDepth = 30, ignore = new Set(), childOf = new opentracing_1.Span()) {
        const span = childOf.tracer().startSpan('Ensure referenced files', { childOf });
        span.addTags({ uri, maxDepth });
        ignore.add(uri);
        return rxjs_1.Observable.from(this.ensureModuleStructure(span))
            .mergeMap(() => maxDepth === 0 ? [] : this.resolveReferencedFiles(uri))
            .filter(referencedUri => !ignore.has(referencedUri))
            .mergeMap(referencedUri => this.ensureReferencedFiles(referencedUri, maxDepth - 1, ignore)
            .catch(err => {
            this.logger.error(`Error resolving file references for ${uri}:`, err);
            return [];
        }))
            .catch(err => {
            span.setTag('error', true);
            span.log({ 'event': 'error', 'error.object': err, 'message': err.message, 'stack': err.stack });
            throw err;
        })
            .finally(() => {
            span.finish();
        });
    }
    /**
     * Invalidates a cache entry for `resolveReferencedFiles` (e.g. because the file changed)
     *
     * @param uri The URI that referenced files should be invalidated for. If not given, all entries are invalidated
     */
    invalidateReferencedFiles(uri) {
        if (uri) {
            this.referencedFiles.delete(uri);
        }
        else {
            this.referencedFiles.clear();
        }
    }
    /**
     * Returns the files that are referenced from a given file.
     * If the file has already been processed, returns a cached value.
     *
     * @param uri URI of the file to process
     * @return URIs of files referenced by the file
     */
    resolveReferencedFiles(uri) {
        let observable = this.referencedFiles.get(uri);
        if (observable) {
            return observable;
        }
        const parts = url.parse(uri);
        if (!parts.pathname) {
            return rxjs_1.Observable.throw(new Error(`Invalid URI ${uri}`));
        }
        // TypeScript works with file paths, not URIs
        const filePath = parts.pathname.split('/').map(decodeURIComponent).join('/');
        observable = rxjs_1.Observable.from(this.updater.ensure(uri))
            .mergeMap(() => {
            const config = this.getConfiguration(filePath);
            config.ensureBasicFiles();
            const contents = this.localFs.getContent(uri);
            const info = ts.preProcessFile(contents, true, true);
            const compilerOpt = config.getHost().getCompilationSettings();
            // TODO remove platform-specific behavior here, the host OS is not coupled to the client OS
            const resolver = !this.strict && os.platform() === 'win32' ? path_ : path_.posix;
            // Iterate imported files
            return rxjs_1.Observable.merge(
            // References with `import`
            rxjs_1.Observable.from(info.importedFiles)
                .map(importedFile => ts.resolveModuleName(util.toUnixPath(importedFile.fileName), filePath, compilerOpt, config.moduleResolutionHost()))
                .filter(resolved => !!(resolved && resolved.resolvedModule))
                .map(resolved => resolved.resolvedModule.resolvedFileName), 
            // References with `<reference path="..."/>`
            rxjs_1.Observable.from(info.referencedFiles)
                .map(referencedFile => util.toUnixPath(path_.relative(this.rootPath, resolver.resolve(this.rootPath, resolver.dirname(filePath), util.toUnixPath(referencedFile.fileName))))));
        })
            .map(filePath => url.format(Object.assign({}, parts, { pathname: filePath.split(/[\\\/]/).map(encodeURIComponent).join('/'), search: undefined, hash: undefined })))
            .catch(err => {
            this.referencedFiles.delete(uri);
            throw err;
        })
            .publishReplay()
            .refCount();
        this.referencedFiles.set(uri, observable);
        return observable;
    }
    /**
     * @param filePath source file path relative to project root
     * @return project configuration for a given source file. Climbs directory tree up to workspace root if needed
     */
    getConfiguration(filePath) {
        const config = this.getConfigurationIfExists(filePath);
        if (!config) {
            throw new Error(`TypeScript config file for ${filePath} not found`);
        }
        return config;
    }
    /**
     * @param filePath source file path relative to project root
     * @return closest configuration for a given file path or undefined if there is no such configuration
     */
    getConfigurationIfExists(filePath) {
        let dir = filePath;
        let config;
        while (dir && dir !== this.rootPath) {
            config = this.configs.get(dir);
            if (config) {
                return config;
            }
            dir = path_.posix.dirname(dir);
            if (dir === '.') {
                dir = '';
            }
        }
        return this.configs.get('');
    }
    /**
     * Called when file was opened by client. Current implementation
     * does not differenciates open and change events
     * @param filePath path to a file relative to project root
     * @param text file's content
     */
    didOpen(filePath, text) {
        this.didChange(filePath, text);
    }
    /**
     * Called when file was closed by client. Current implementation invalidates compiled version
     * @param filePath path to a file relative to project root
     */
    didClose(filePath) {
        this.localFs.didClose(filePath);
        let version = this.versions.get(filePath) || 0;
        this.versions.set(filePath, ++version);
        const config = this.getConfigurationIfExists(filePath);
        if (!config) {
            return;
        }
        config.ensureConfigFile();
        config.getHost().incProjectVersion();
        config.syncProgram();
    }
    /**
     * Called when file was changed by client. Current implementation invalidates compiled version
     * @param filePath path to a file relative to project root
     * @param text file's content
     */
    didChange(filePath, text) {
        this.localFs.didChange(filePath, text);
        let version = this.versions.get(filePath) || 0;
        this.versions.set(filePath, ++version);
        const config = this.getConfigurationIfExists(filePath);
        if (!config) {
            return;
        }
        config.ensureConfigFile();
        config.getHost().incProjectVersion();
        config.syncProgram();
    }
    /**
     * Called when file was saved by client
     * @param filePath path to a file relative to project root
     */
    didSave(filePath) {
        this.localFs.didSave(filePath);
    }
    /**
     * Detects projects and creates projects denoted by tsconfig.json and jsconfig.json fiels.
     * Previously detected projects are NOT discarded.
     * If there is no root configuration, adds it to catch all orphan files
     */
    createConfigurations() {
        const rootdirs = new Set();
        for (const uri of this.localFs.uris()) {
            const relativeFilePath = path_.posix.relative(this.rootPath, util.uri2path(uri));
            if (!/(^|\/)[tj]sconfig\.json$/.test(relativeFilePath)) {
                continue;
            }
            if (/(^|\/)node_modules\//.test(relativeFilePath)) {
                continue;
            }
            let dir = path_.posix.dirname(relativeFilePath);
            if (dir === '.') {
                dir = '';
            }
            if (!this.configs.has(dir)) {
                this.configs.set(dir, new ProjectConfiguration(this.localFs, path_.posix.join('/', dir), this.versions, relativeFilePath, undefined, this.traceModuleResolution, this.logger));
            }
            rootdirs.add(dir);
        }
        if (!rootdirs.has('') && !this.configs.has('')) {
            // collecting all the files in workspace by making fake configuration object
            this.configs.set('', new ProjectConfiguration(this.localFs, '/', this.versions, '', {
                compilerOptions: {
                    module: ts.ModuleKind.CommonJS,
                    allowNonTsExtensions: false,
                    allowJs: true
                }
            }, this.traceModuleResolution, this.logger));
        }
    }
}
exports.ProjectManager = ProjectManager;
/**
 * Implementaton of LanguageServiceHost that works with in-memory file system.
 * It takes file content from local cache and provides it to TS compiler on demand
 *
 * @implements ts.LanguageServiceHost
 */
class InMemoryLanguageServiceHost {
    constructor(rootPath, options, fs, expectedFiles, versions, logger = new logging_1.NoopLogger()) {
        this.logger = logger;
        this.rootPath = rootPath;
        this.options = options;
        this.fs = fs;
        this.expectedFilePaths = expectedFiles;
        this.versions = versions;
        this.projectVersion = 1;
        this.filePaths = [];
    }
    /**
     * TypeScript uses this method (when present) to compare project's version
     * with the last known one to decide if internal data should be synchronized
     */
    getProjectVersion() {
        return '' + this.projectVersion;
    }
    /**
     * Incrementing current project version, telling TS compiler to invalidate internal data
     */
    incProjectVersion() {
        this.projectVersion++;
    }
    getCompilationSettings() {
        return this.options;
    }
    getScriptFileNames() {
        return this.filePaths;
    }
    /**
     * Adds a file and increments project version, used in conjunction with getProjectVersion()
     * which may be called by TypeScript to check if internal data is up to date
     *
     * @param filePath relative file path
     */
    addFile(filePath) {
        this.filePaths.push(filePath);
        this.incProjectVersion();
    }
    /**
     * @param fileName relative or absolute file path
     */
    getScriptVersion(fileName) {
        if (path_.posix.isAbsolute(fileName) || path_.isAbsolute(fileName)) {
            fileName = path_.posix.relative(this.rootPath, util.toUnixPath(fileName));
        }
        let version = this.versions.get(fileName);
        if (!version) {
            version = 1;
            this.versions.set(fileName, version);
        }
        return '' + version;
    }
    /**
     * @param fileName relative or absolute file path
     */
    getScriptSnapshot(fileName) {
        let exists = this.fs.fileExists(fileName);
        if (!exists) {
            fileName = path_.posix.join(this.rootPath, fileName);
            exists = this.fs.fileExists(fileName);
        }
        if (!exists) {
            return undefined;
        }
        return ts.ScriptSnapshot.fromString(this.fs.readFile(fileName));
    }
    getCurrentDirectory() {
        return this.rootPath;
    }
    getDefaultLibFileName(options) {
        return util.toUnixPath(ts.getDefaultLibFilePath(options));
    }
    trace(message) {
        // empty
    }
    log(message) {
        // empty
    }
    error(message) {
        this.logger.error(message);
    }
}
exports.InMemoryLanguageServiceHost = InMemoryLanguageServiceHost;
/**
 * ProjectConfiguration instances track the compiler configuration (as
 * defined by {tj}sconfig.json if it exists) and state for a single
 * TypeScript project. It represents the world of the view as
 * presented to the compiler.
 *
 * For efficiency, a ProjectConfiguration instance may hide some files
 * from the compiler, preventing them from being parsed and
 * type-checked. Depending on the use, the caller should call one of
 * the ensure* methods to ensure that the appropriate files have been
 * made available to the compiler before calling any other methods on
 * the ProjectConfiguration or its public members. By default, no
 * files are parsed.
 */
class ProjectConfiguration {
    /**
     * @param fs file system to use
     * @param rootFilePath root file path, relative to workspace hierarchy root
     * @param configFilePath configuration file path (relative to workspace root)
     * @param configContent optional configuration content to use instead of reading configuration file)
     */
    constructor(fs, rootFilePath, versions, configFilePath, configContent, traceModuleResolution, logger = new logging_1.NoopLogger()) {
        this.logger = logger;
        this.initialized = false;
        this.ensuredBasicFiles = false;
        this.ensuredAllFiles = false;
        this.fs = fs;
        this.configFilePath = configFilePath;
        this.configContent = configContent;
        this.versions = versions;
        this.traceModuleResolution = traceModuleResolution || false;
        this.rootFilePath = rootFilePath;
    }
    /**
     * @return module resolution host to use by TS service
     */
    moduleResolutionHost() {
        return this.fs;
    }
    /**
     * reset resets a ProjectConfiguration to its state immediately
     * after construction. It should be called whenever the underlying
     * local filesystem (fs) has changed, and so the
     * ProjectConfiguration can no longer assume its state reflects
     * that of the underlying files.
     */
    reset() {
        this.initialized = false;
        this.ensuredBasicFiles = false;
        this.ensuredAllFiles = false;
        this.service = undefined;
        this.program = undefined;
        this.host = undefined;
    }
    /**
     * @return package name (project name) of a given project
     */
    getPackageName() {
        // package.json may be located at the upper level as well
        let currentDir = this.rootFilePath;
        while (true) {
            const pkgJsonFile = path_.posix.join(currentDir, 'package.json');
            if (this.fs.fileExists(pkgJsonFile)) {
                return JSON.parse(this.fs.readFile(pkgJsonFile)).name;
            }
            const parentDir = path_.dirname(currentDir);
            if (parentDir === '.' || parentDir === '/' || parentDir === currentDir) {
                return undefined;
            }
            currentDir = parentDir;
        }
    }
    /**
     * @return language service object
     */
    getService() {
        if (!this.service) {
            throw new Error('project is uninitialized');
        }
        return this.service;
    }
    /**
     * Note that it does not perform any parsing or typechecking
     * @return program object (cached result of parsing and typechecking done by TS service)
     */
    getProgram() {
        if (!this.program) {
            throw new Error('project is uninitialized');
        }
        return this.program;
    }
    /**
     * @return language service host that TS service uses to read the data
     */
    getHost() {
        if (!this.host) {
            throw new Error('project is uninitialized');
        }
        return this.host;
    }
    /**
     * Tells TS service to recompile program (if needed) based on current list of files and compilation options.
     * TS service relies on information provided by language servide host to see if there were any changes in
     * the whole project or in some files
     */
    syncProgram() {
        this.program = this.getService().getProgram();
    }
    /**
     * Initializes (sub)project by parsing configuration and making proper internal objects
     */
    init() {
        if (this.initialized) {
            return;
        }
        let configObject;
        if (!this.configContent) {
            const jsonConfig = ts.parseConfigFileTextToJson(this.configFilePath, this.fs.readFile(this.configFilePath));
            if (jsonConfig.error) {
                this.logger.error('Cannot parse ' + this.configFilePath + ': ' + jsonConfig.error.messageText);
                throw new Error('Cannot parse ' + this.configFilePath + ': ' + jsonConfig.error.messageText);
            }
            configObject = jsonConfig.config;
        }
        else {
            configObject = this.configContent;
        }
        let dir = path_.posix.dirname(this.configFilePath);
        if (dir === '.') {
            dir = '';
        }
        const base = dir || this.fs.path;
        const configParseResult = ts.parseJsonConfigFileContent(configObject, this.fs, base);
        const expFiles = configParseResult.fileNames;
        // Add globals that might exist in dependencies
        const nodeModulesDir = path_.posix.join(base, 'node_modules');
        const err = memfs_1.walkInMemoryFs(this.fs, nodeModulesDir, (path, isdir) => {
            if (!isdir && util.isGlobalTSFile(path)) {
                expFiles.push(path);
            }
        });
        if (err) {
            throw err;
        }
        const options = configParseResult.options;
        if (/(^|\/)jsconfig\.json$/.test(this.configFilePath)) {
            options.allowJs = true;
        }
        if (this.traceModuleResolution) {
            options.traceResolution = true;
        }
        this.host = new InMemoryLanguageServiceHost(this.fs.path, options, this.fs, expFiles, this.versions, this.logger);
        this.service = ts.createLanguageService(this.host, ts.createDocumentRegistry());
        this.program = this.service.getProgram();
        this.initialized = true;
    }
    /**
     * Ensures we are ready to process files from a given sub-project
     */
    ensureConfigFile() {
        this.init();
    }
    /**
     * Ensures we added basic files (global TS files, dependencies, declarations)
     */
    ensureBasicFiles() {
        if (this.ensuredBasicFiles) {
            return;
        }
        this.init();
        let changed = false;
        for (const fileName of (this.getHost().expectedFilePaths || [])) {
            if (util.isGlobalTSFile(fileName) || (!util.isDependencyFile(fileName) && util.isDeclarationFile(fileName))) {
                const sourceFile = this.getProgram().getSourceFile(fileName);
                if (!sourceFile) {
                    this.getHost().addFile(fileName);
                    changed = true;
                }
            }
        }
        if (changed) {
            // requery program object to synchonize LanguageService's data
            this.program = this.getService().getProgram();
        }
        this.ensuredBasicFiles = true;
    }
    /**
     * Ensures we added all project's source file (as were defined in tsconfig.json)
     */
    ensureAllFiles() {
        if (this.ensuredAllFiles) {
            return;
        }
        this.init();
        if (this.getHost().complete) {
            return;
        }
        let changed = false;
        for (const fileName of (this.getHost().expectedFilePaths || [])) {
            const sourceFile = this.getProgram().getSourceFile(fileName);
            if (!sourceFile) {
                this.getHost().addFile(fileName);
                changed = true;
            }
        }
        if (changed) {
            // requery program object to synchonize LanguageService's data
            this.program = this.getService().getProgram();
        }
        this.getHost().complete = true;
        this.ensuredAllFiles = true;
    }
}
exports.ProjectConfiguration = ProjectConfiguration;
//# sourceMappingURL=project-manager.js.map