import { Observable } from '@reactivex/rxjs';
import { Span } from 'opentracing';
import { CompletionList, DidChangeTextDocumentParams, DidCloseTextDocumentParams, DidOpenTextDocumentParams, DidSaveTextDocumentParams, DocumentSymbolParams, Hover, InitializeResult, Location, ReferenceParams, SignatureHelp, SymbolInformation, TextDocumentPositionParams } from 'vscode-languageserver';
import { FileSystem, FileSystemUpdater } from './fs';
import { LanguageClient } from './lang-handler';
import { Logger } from './logging';
import { InMemoryFileSystem } from './memfs';
import * as pm from './project-manager';
import { DependencyReference, InitializeParams, PackageInformation, ReferenceInformation, SymbolLocationInformation, WorkspaceReferenceParams, WorkspaceSymbolParams } from './request-type';
export interface TypeScriptServiceOptions {
    traceModuleResolution?: boolean;
    strict?: boolean;
}
export declare type TypeScriptServiceFactory = (client: LanguageClient, options?: TypeScriptServiceOptions) => TypeScriptService;
/**
 * Handles incoming requests and return responses. There is a one-to-one-to-one
 * correspondence between TCP connection, TypeScriptService instance, and
 * language workspace. TypeScriptService caches data from the compiler across
 * requests. The lifetime of the TypeScriptService instance is tied to the
 * lifetime of the TCP connection, so its caches are deleted after the
 * connection is torn down.
 *
 * Methods are camelCase versions of the LSP spec methods and dynamically
 * dispatched. Methods not to be exposed over JSON RPC are prefixed with an
 * underscore.
 */
export declare class TypeScriptService {
    protected client: LanguageClient;
    protected options: TypeScriptServiceOptions;
    projectManager: pm.ProjectManager;
    /**
     * The rootPath as passed to `initialize` or converted from `rootUri`
     */
    root: string;
    /**
     * The root URI as passed to `initialize` or converted from `rootPath`
     */
    protected rootUri: string;
    /**
     * Cached response for empty workspace/symbol query
     */
    private emptyQueryWorkspaceSymbols;
    private traceModuleResolution;
    /**
     * The remote (or local), asynchronous, file system to fetch files from
     */
    protected fileSystem: FileSystem;
    protected logger: Logger;
    /**
     * Holds file contents and workspace structure in memory
     */
    protected inMemoryFileSystem: InMemoryFileSystem;
    /**
     * Syncs the remote file system with the in-memory file system
     */
    protected updater: FileSystemUpdater;
    constructor(client: LanguageClient, options?: TypeScriptServiceOptions);
    /**
     * The initialize request is sent as the first request from the client to the server. If the
     * server receives request or notification before the `initialize` request it should act as
     * follows:
     *
     * - for a request the respond should be errored with `code: -32002`. The message can be picked by
     * the server.
     * - notifications should be dropped, except for the exit notification. This will allow the exit a
     * server without an initialize request.
     *
     * Until the server has responded to the `initialize` request with an `InitializeResult` the
     * client must not sent any additional requests or notifications to the server.
     *
     * During the `initialize` request the server is allowed to sent the notifications
     * `window/showMessage`, `window/logMessage` and `telemetry/event` as well as the
     * `window/showMessageRequest` request to the client.
     */
    initialize(params: InitializeParams, span?: Span): Promise<InitializeResult>;
    /**
     * Initializes the remote file system and in-memory file system.
     * Can be overridden
     *
     * @param accessDisk Whether the language server is allowed to access the local file system
     */
    protected _initializeFileSystems(accessDisk: boolean): void;
    /**
     * The shutdown request is sent from the client to the server. It asks the server to shut down,
     * but to not exit (otherwise the response might not be delivered correctly to the client).
     * There is a separate exit notification that asks the server to exit.
     */
    shutdown(params?: {}, span?: Span): Promise<null>;
    /**
     * The goto definition request is sent from the client to the server to resolve the definition
     * location of a symbol at a given text document position.
     */
    textDocumentDefinition(params: TextDocumentPositionParams, span?: Span): Promise<Location[]>;
    textDocumentXdefinition(params: TextDocumentPositionParams, span?: Span): Promise<SymbolLocationInformation[]>;
    /**
     * The hover request is sent from the client to the server to request hover information at a
     * given text document position.
     */
    textDocumentHover(params: TextDocumentPositionParams, span?: Span): Promise<Hover>;
    /**
     * The references request is sent from the client to the server to resolve project-wide
     * references for the symbol denoted by the given text document position.
     */
    textDocumentReferences(params: ReferenceParams, span?: Span): Promise<Location[]>;
    /**
     * The workspace symbol request is sent from the client to the server to list project-wide
     * symbols matching the query string. The text document parameter specifies the active document
     * at time of the query. This can be used to rank or limit results.
     */
    workspaceSymbol(params: WorkspaceSymbolParams, span?: Span): Promise<SymbolInformation[]>;
    protected _workspaceSymbolDefinitelyTyped(params: WorkspaceSymbolParams): Promise<SymbolInformation[] | null>;
    /**
     * The document symbol request is sent from the client to the server to list all symbols found
     * in a given text document.
     */
    textDocumentDocumentSymbol(params: DocumentSymbolParams, span?: Span): Promise<SymbolInformation[]>;
    /**
     * The workspace references request is sent from the client to the server to locate project-wide
     * references to a symbol given its description / metadata.
     */
    workspaceXreferences(params: WorkspaceReferenceParams, span?: Span): Promise<ReferenceInformation[]>;
    /**
     * This method returns metadata about the package(s) defined in a workspace and a list of
     * dependencies for each package.
     *
     * This method is necessary to implement cross-repository jump-to-def when it is not possible to
     * resolve the global location of the definition from data present or derived from the local
     * workspace. For example, a package manager might not include information about the source
     * repository of each dependency. In this case, definition resolution requires mapping from
     * package descriptor to repository revision URL. A reverse index can be constructed from calls
     * to workspace/xpackages to provide an efficient mapping.
     */
    workspaceXpackages(params?: {}, span?: Span): Observable<PackageInformation[]>;
    /**
     * Returns all dependencies of a workspace.
     * Superseded by workspace/xpackages
     */
    workspaceXdependencies(params?: {}, span?: Span): Observable<DependencyReference[]>;
    /**
     * The Completion request is sent from the client to the server to compute completion items at a
     * given cursor position. Completion items are presented in the
     * [IntelliSense](https://code.visualstudio.com/docs/editor/editingevolved#_intellisense) user
     * interface. If computing full completion items is expensive, servers can additionally provide
     * a handler for the completion item resolve request ('completionItem/resolve'). This request is
     * sent when a completion item is selected in the user interface. A typically use case is for
     * example: the 'textDocument/completion' request doesn't fill in the `documentation` property
     * for returned completion items since it is expensive to compute. When the item is selected in
     * the user interface then a 'completionItem/resolve' request is sent with the selected
     * completion item as a param. The returned completion item should have the documentation
     * property filled in.
     */
    textDocumentCompletion(params: TextDocumentPositionParams, span?: Span): Promise<CompletionList>;
    /**
     * The signature help request is sent from the client to the server to request signature
     * information at a given cursor position.
     */
    textDocumentSignatureHelp(params: TextDocumentPositionParams, span?: Span): Promise<SignatureHelp>;
    private _walkMostAST(node, visit);
    /**
     * The document open notification is sent from the client to the server to signal newly opened
     * text documents. The document's truth is now managed by the client and the server must not try
     * to read the document's truth using the document's uri.
     */
    textDocumentDidOpen(params: DidOpenTextDocumentParams): Promise<void>;
    /**
     * The document change notification is sent from the client to the server to signal changes to a
     * text document. In 2.0 the shape of the params has changed to include proper version numbers
     * and language ids.
     */
    textDocumentDidChange(params: DidChangeTextDocumentParams): Promise<void>;
    /**
     * The document save notification is sent from the client to the server when the document was
     * saved in the client.
     */
    textDocumentDidSave(params: DidSaveTextDocumentParams): Promise<void>;
    /**
     * The document close notification is sent from the client to the server when the document got
     * closed in the client. The document's truth now exists where the document's uri points to
     * (e.g. if the document's uri is a file uri the truth now exists on disk).
     */
    textDocumentDidClose(params: DidCloseTextDocumentParams): Promise<void>;
    /**
     * Fetches (or creates if needed) source file object for a given file name
     * @param configuration project configuration
     * @param fileName file name to fetch source file for or create it
     */
    private _getSourceFile(configuration, fileName);
    /**
     * Produces async function that converts ReferenceEntry object to Location
     */
    private _transformReference(root, program, includeDeclaration, ref);
    /**
     * transformNavItem transforms a NavigateToItem instance to a SymbolInformation instance
     */
    private _transformNavItem(program, item);
    /**
     * Returns an Iterator for all symbols in a given config
     *
     * Note: This method is not traced because it returns an Iterator that may produce values lazily
     *
     * @param config The ProjectConfiguration to search
     * @param query A text or SymbolDescriptor query
     * @param limit An optional limit that is passed to TypeScript
     * @return Iterator that emits SymbolInformations
     */
    private _collectWorkspaceSymbols(config, query?, limit?);
    /**
     * Transforms definition's file name to URI. If definition belongs to TypeScript library,
     * returns git://github.com/Microsoft/TypeScript URL, otherwise returns file:// one
     */
    private _defUri(filePath);
    /**
     * Fetches up to limit navigation bar items from given project, flattens them
     */
    private _getNavigationTreeItems(configuration);
    /**
     * Flattens navigation tree by emitting acceptable NavigationTreeItems as SymbolInformations.
     * Some items (source files, modules) may be excluded
     */
    private _flattenNavigationTreeItem(item, parent, sourceFile);
    /**
     * Transforms NavigationTree to SymbolInformation
     */
    private _transformNavigationTreeItem(item, parent, sourceFile);
    /**
     * @return true if navigation tree item is acceptable for inclusion into workspace/symbols
     */
    private static isAcceptableNavigationTreeItem(item);
}
