"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const rxjs_1 = require("@reactivex/rxjs");
const async = require("async");
const iterare_1 = require("iterare");
const lodash_1 = require("lodash");
const opentracing_1 = require("opentracing");
const path_ = require("path");
const ts = require("typescript");
const url = require("url");
const vscode_languageserver_1 = require("vscode-languageserver");
const fs_1 = require("./fs");
const logging_1 = require("./logging");
const memfs_1 = require("./memfs");
const pm = require("./project-manager");
const util = require("./util");
/**
 * Handles incoming requests and return responses. There is a one-to-one-to-one
 * correspondence between TCP connection, TypeScriptService instance, and
 * language workspace. TypeScriptService caches data from the compiler across
 * requests. The lifetime of the TypeScriptService instance is tied to the
 * lifetime of the TCP connection, so its caches are deleted after the
 * connection is torn down.
 *
 * Methods are camelCase versions of the LSP spec methods and dynamically
 * dispatched. Methods not to be exposed over JSON RPC are prefixed with an
 * underscore.
 */
class TypeScriptService {
    constructor(client, options = {}) {
        this.client = client;
        this.options = options;
        this.logger = new logging_1.LSPLogger(client);
    }
    /**
     * The initialize request is sent as the first request from the client to the server. If the
     * server receives request or notification before the `initialize` request it should act as
     * follows:
     *
     * - for a request the respond should be errored with `code: -32002`. The message can be picked by
     * the server.
     * - notifications should be dropped, except for the exit notification. This will allow the exit a
     * server without an initialize request.
     *
     * Until the server has responded to the `initialize` request with an `InitializeResult` the
     * client must not sent any additional requests or notifications to the server.
     *
     * During the `initialize` request the server is allowed to sent the notifications
     * `window/showMessage`, `window/logMessage` and `telemetry/event` as well as the
     * `window/showMessageRequest` request to the client.
     */
    initialize(params, span = new opentracing_1.Span()) {
        return __awaiter(this, void 0, void 0, function* () {
            if (params.rootUri || params.rootPath) {
                this.root = params.rootPath || util.uri2path(params.rootUri);
                this.rootUri = params.rootUri || util.path2uri('', params.rootPath);
                this._initializeFileSystems(!this.options.strict && !(params.capabilities.xcontentProvider && params.capabilities.xfilesProvider));
                this.updater = new fs_1.FileSystemUpdater(this.fileSystem, this.inMemoryFileSystem);
                this.projectManager = new pm.ProjectManager(this.root, this.inMemoryFileSystem, this.updater, !!this.options.strict, this.traceModuleResolution, this.logger);
                // Pre-fetch files in the background
                this.projectManager.ensureOwnFiles(span).catch(err => {
                    this.logger.error('Background fetching failed ', err);
                });
            }
            return {
                capabilities: {
                    // Tell the client that the server works in FULL text document sync mode
                    textDocumentSync: vscode_languageserver_1.TextDocumentSyncKind.Full,
                    hoverProvider: true,
                    signatureHelpProvider: {
                        triggerCharacters: ['(', ',']
                    },
                    definitionProvider: true,
                    referencesProvider: true,
                    documentSymbolProvider: true,
                    workspaceSymbolProvider: true,
                    xworkspaceReferencesProvider: true,
                    xdefinitionProvider: true,
                    xdependenciesProvider: true,
                    completionProvider: {
                        resolveProvider: false,
                        triggerCharacters: ['.']
                    },
                    xpackagesProvider: true
                }
            };
        });
    }
    /**
     * Initializes the remote file system and in-memory file system.
     * Can be overridden
     *
     * @param accessDisk Whether the language server is allowed to access the local file system
     */
    _initializeFileSystems(accessDisk) {
        this.fileSystem = accessDisk ? new fs_1.LocalFileSystem(util.uri2path(this.root)) : new fs_1.RemoteFileSystem(this.client);
        this.inMemoryFileSystem = new memfs_1.InMemoryFileSystem(this.root);
    }
    /**
     * The shutdown request is sent from the client to the server. It asks the server to shut down,
     * but to not exit (otherwise the response might not be delivered correctly to the client).
     * There is a separate exit notification that asks the server to exit.
     */
    shutdown(params = {}, span = new opentracing_1.Span()) {
        return __awaiter(this, void 0, void 0, function* () {
            this.projectManager.dispose();
            return null;
        });
    }
    /**
     * The goto definition request is sent from the client to the server to resolve the definition
     * location of a symbol at a given text document position.
     */
    textDocumentDefinition(params, span = new opentracing_1.Span()) {
        return __awaiter(this, void 0, void 0, function* () {
            const uri = util.uri2reluri(params.textDocument.uri, this.root);
            const line = params.position.line;
            const column = params.position.character;
            // Fetch files needed to resolve definition
            yield this.projectManager.ensureReferencedFiles(params.textDocument.uri, undefined, undefined, span).toPromise();
            const fileName = util.uri2path(uri);
            const configuration = this.projectManager.getConfiguration(fileName);
            configuration.ensureBasicFiles();
            const sourceFile = this._getSourceFile(configuration, fileName);
            if (!sourceFile) {
                return [];
            }
            const offset = ts.getPositionOfLineAndCharacter(sourceFile, line, column);
            const defs = configuration.getService().getDefinitionAtPosition(fileName, offset);
            const ret = [];
            if (defs) {
                for (const def of defs) {
                    const sourceFile = this._getSourceFile(configuration, def.fileName);
                    if (!sourceFile) {
                        throw new Error('expected source file "' + def.fileName + '" to exist in configuration');
                    }
                    const start = ts.getLineAndCharacterOfPosition(sourceFile, def.textSpan.start);
                    const end = ts.getLineAndCharacterOfPosition(sourceFile, def.textSpan.start + def.textSpan.length);
                    ret.push(vscode_languageserver_1.Location.create(this._defUri(def.fileName), {
                        start,
                        end
                    }));
                }
            }
            return ret;
        });
    }
    textDocumentXdefinition(params, span = new opentracing_1.Span()) {
        return __awaiter(this, void 0, void 0, function* () {
            // Ensure files needed to resolve SymbolLocationInformation are fetched
            yield this.projectManager.ensureReferencedFiles(params.textDocument.uri, undefined, undefined, span).toPromise();
            const fileName = util.uri2path(params.textDocument.uri);
            const configuration = this.projectManager.getConfiguration(fileName);
            configuration.ensureBasicFiles();
            const sourceFile = this._getSourceFile(configuration, fileName);
            if (!sourceFile) {
                return [];
            }
            const offset = ts.getPositionOfLineAndCharacter(sourceFile, params.position.line, params.position.character);
            const defs = configuration.getService().getDefinitionAtPosition(fileName, offset);
            const ret = [];
            if (defs) {
                for (const def of defs) {
                    const sourceFile = this._getSourceFile(configuration, def.fileName);
                    if (!sourceFile) {
                        throw new Error('expected source file "' + def.fileName + '" to exist in configuration');
                    }
                    const start = ts.getLineAndCharacterOfPosition(sourceFile, def.textSpan.start);
                    const end = ts.getLineAndCharacterOfPosition(sourceFile, def.textSpan.start + def.textSpan.length);
                    const loc = vscode_languageserver_1.Location.create(this._defUri(def.fileName), {
                        start,
                        end
                    });
                    ret.push({
                        symbol: util.defInfoToSymbolDescriptor(def),
                        location: loc
                    });
                }
            }
            return ret;
        });
    }
    /**
     * The hover request is sent from the client to the server to request hover information at a
     * given text document position.
     */
    textDocumentHover(params, span = new opentracing_1.Span()) {
        return __awaiter(this, void 0, void 0, function* () {
            // Ensure files needed to resolve hover are fetched
            yield this.projectManager.ensureReferencedFiles(params.textDocument.uri, undefined, undefined, span).toPromise();
            const fileName = util.uri2path(params.textDocument.uri);
            const configuration = this.projectManager.getConfiguration(fileName);
            configuration.ensureBasicFiles();
            const sourceFile = this._getSourceFile(configuration, fileName);
            if (!sourceFile) {
                throw new Error(`Unknown text document ${params.textDocument.uri}`);
            }
            const offset = ts.getPositionOfLineAndCharacter(sourceFile, params.position.line, params.position.character);
            const info = configuration.getService().getQuickInfoAtPosition(fileName, offset);
            if (!info) {
                return { contents: [] };
            }
            const contents = [{
                    language: 'typescript',
                    value: ts.displayPartsToString(info.displayParts)
                }];
            const documentation = ts.displayPartsToString(info.documentation);
            if (documentation) {
                contents.push(documentation);
            }
            const start = ts.getLineAndCharacterOfPosition(sourceFile, info.textSpan.start);
            const end = ts.getLineAndCharacterOfPosition(sourceFile, info.textSpan.start + info.textSpan.length);
            return { contents, range: vscode_languageserver_1.Range.create(start.line, start.character, end.line, end.character) };
        });
    }
    /**
     * The references request is sent from the client to the server to resolve project-wide
     * references for the symbol denoted by the given text document position.
     */
    textDocumentReferences(params, span = new opentracing_1.Span()) {
        return __awaiter(this, void 0, void 0, function* () {
            // Ensure all files were fetched to collect all references
            yield this.projectManager.ensureAllFiles(span);
            const fileName = util.uri2path(params.textDocument.uri);
            const configuration = this.projectManager.getConfiguration(fileName);
            configuration.ensureAllFiles();
            const sourceFile = this._getSourceFile(configuration, fileName);
            if (!sourceFile) {
                return [];
            }
            const offset = ts.getPositionOfLineAndCharacter(sourceFile, params.position.line, params.position.character);
            const refs = configuration.getService().getReferencesAtPosition(fileName, offset);
            const tasks = [];
            if (refs) {
                for (const ref of refs) {
                    tasks.push(this._transformReference(this.root, configuration.getProgram(), params.context && params.context.includeDeclaration, ref));
                }
            }
            return new Promise((resolve, reject) => {
                async.parallel(tasks, (err, results) => {
                    return resolve(results.filter(item => item));
                });
            });
        });
    }
    /**
     * The workspace symbol request is sent from the client to the server to list project-wide
     * symbols matching the query string. The text document parameter specifies the active document
     * at time of the query. This can be used to rank or limit results.
     */
    workspaceSymbol(params, span = new opentracing_1.Span()) {
        return __awaiter(this, void 0, void 0, function* () {
            // Always return max. 50 results
            // TODO stream 50 results, then re-query and stream the rest
            const limit = Math.min(params.limit || Infinity, 50);
            const query = params.query;
            const symbolQuery = params.symbol ? Object.assign({}, params.symbol) : undefined;
            if (symbolQuery && symbolQuery.package) {
                symbolQuery.package = { name: symbolQuery.package.name };
            }
            if (symbolQuery) {
                try {
                    const dtRes = yield this._workspaceSymbolDefinitelyTyped(Object.assign({}, params, { limit }));
                    if (dtRes) {
                        return dtRes;
                    }
                }
                catch (err) {
                    // Ignore
                }
                if (!symbolQuery.containerKind) {
                    // symbolQuery.containerKind is sometimes empty when symbol.containerKind = 'module'
                    symbolQuery.containerKind = undefined;
                }
            }
            // A workspace/symol request searches all symbols in own code, but not in dependencies
            yield this.projectManager.ensureOwnFiles(span);
            // Cache result for empty query
            if (!query && !symbolQuery && this.emptyQueryWorkspaceSymbols) {
                return this.emptyQueryWorkspaceSymbols;
            }
            // Find configurations to search
            let configs;
            if (symbolQuery && symbolQuery.package && symbolQuery.package.name) {
                // If PackageDescriptor is given, only search project with the matching package name
                configs = iterare_1.default(this.projectManager.configurations())
                    .filter(config => config.getPackageName() === symbolQuery.package.name);
            }
            else {
                const rootConfig = this.projectManager.getConfiguration('');
                if (rootConfig) {
                    // Use root configuration because it includes all files
                    configs = [rootConfig];
                }
                else {
                    // Use all configurations
                    configs = this.projectManager.configurations();
                }
            }
            const symbols = iterare_1.default(configs)
                .map(config => this._collectWorkspaceSymbols(config, query || symbolQuery, limit))
                .flatten()
                .take(limit)
                .toArray();
            // Save empty query result
            if (!query && !symbolQuery) {
                this.emptyQueryWorkspaceSymbols = symbols;
            }
            return symbols;
        });
    }
    _workspaceSymbolDefinitelyTyped(params) {
        return __awaiter(this, void 0, void 0, function* () {
            const rootUriParts = url.parse(this.rootUri);
            if (!rootUriParts.pathname) {
                return null;
            }
            const packageJsonUri = url.format(Object.assign({}, rootUriParts, { pathname: path_.posix.join(rootUriParts.pathname, 'package.json') }));
            yield this.updater.ensure(packageJsonUri);
            const rootConfig = JSON.parse(this.inMemoryFileSystem.getContent(packageJsonUri));
            if (rootConfig.name !== 'definitely-typed') {
                return null;
            }
            if (!params.symbol || !params.symbol.package) {
                return null;
            }
            const pkg = params.symbol.package;
            if (!pkg.name || !pkg.name.startsWith('@types/')) {
                return null;
            }
            const relPkgRoot = pkg.name.slice('@types/'.length);
            yield this.projectManager.ensureModuleStructure();
            const symbolQuery = params.symbol ? Object.assign({}, params.symbol) : undefined;
            if (symbolQuery) {
                symbolQuery.package = undefined;
                if (!symbolQuery.containerKind) {
                    symbolQuery.containerKind = undefined; // symQuery.containerKind is sometimes empty when symbol.containerKind = 'module'
                }
            }
            const config = this.projectManager.getConfiguration(relPkgRoot);
            return Array.from(this._collectWorkspaceSymbols(config, params.query || symbolQuery, params.limit));
        });
    }
    /**
     * The document symbol request is sent from the client to the server to list all symbols found
     * in a given text document.
     */
    textDocumentDocumentSymbol(params, span = new opentracing_1.Span()) {
        return __awaiter(this, void 0, void 0, function* () {
            // Ensure files needed to resolve symbols are fetched
            yield this.projectManager.ensureReferencedFiles(params.textDocument.uri, undefined, undefined, span).toPromise();
            const fileName = util.uri2path(params.textDocument.uri);
            const config = this.projectManager.getConfiguration(fileName);
            config.ensureBasicFiles();
            const sourceFile = this._getSourceFile(config, fileName);
            if (!sourceFile) {
                return [];
            }
            const tree = config.getService().getNavigationTree(fileName);
            return Array.from(this._flattenNavigationTreeItem(tree, null, sourceFile));
        });
    }
    /**
     * The workspace references request is sent from the client to the server to locate project-wide
     * references to a symbol given its description / metadata.
     */
    workspaceXreferences(params, span = new opentracing_1.Span()) {
        return __awaiter(this, void 0, void 0, function* () {
            const refInfo = [];
            yield this.projectManager.ensureAllFiles(span);
            const configs = this.projectManager.configurations();
            yield Promise.all(iterare_1.default(configs).map((config) => __awaiter(this, void 0, void 0, function* () {
                if (params.hints && params.hints.dependeePackageName && params.hints.dependeePackageName !== config.getPackageName()) {
                    return;
                }
                config.ensureAllFiles();
                const files = config.getService().getProgram().getSourceFiles().sort((a, b) => a.fileName.localeCompare(b.fileName));
                for (const source of files) {
                    // ignore dependency files
                    if (util.toUnixPath(source.fileName).indexOf(`${path_.posix.sep}node_modules${path_.posix.sep}`) !== -1) {
                        continue;
                    }
                    this._walkMostAST(source, node => {
                        switch (node.kind) {
                            case ts.SyntaxKind.Identifier: {
                                const defs = config.getService().getDefinitionAtPosition(source.fileName, node.pos + 1);
                                if (!defs) {
                                    break;
                                }
                                for (const def of defs) {
                                    const sd = util.defInfoToSymbolDescriptor(def);
                                    if (!util.symbolDescriptorMatch(params.query, sd)) {
                                        continue;
                                    }
                                    const start = ts.getLineAndCharacterOfPosition(source, node.pos);
                                    const end = ts.getLineAndCharacterOfPosition(source, node.end);
                                    const loc = {
                                        uri: this._defUri(source.fileName),
                                        range: {
                                            start,
                                            end
                                        }
                                    };
                                    refInfo.push({
                                        symbol: sd,
                                        reference: loc
                                    });
                                }
                                break;
                            }
                            case ts.SyntaxKind.StringLiteral: {
                                // TODO: include string-interpolated references
                                break;
                            }
                        }
                    });
                }
            })));
            return refInfo;
        });
    }
    /**
     * This method returns metadata about the package(s) defined in a workspace and a list of
     * dependencies for each package.
     *
     * This method is necessary to implement cross-repository jump-to-def when it is not possible to
     * resolve the global location of the definition from data present or derived from the local
     * workspace. For example, a package manager might not include information about the source
     * repository of each dependency. In this case, definition resolution requires mapping from
     * package descriptor to repository revision URL. A reverse index can be constructed from calls
     * to workspace/xpackages to provide an efficient mapping.
     */
    workspaceXpackages(params = {}, span = new opentracing_1.Span()) {
        // Ensure package.json files
        return rxjs_1.Observable.from(this.projectManager.ensureModuleStructure(span))
            .mergeMap(() => this.inMemoryFileSystem.uris())
            .filter(uri => uri.includes('/package.json') && !uri.includes('/node_modules/'))
            .mergeMap(uri => rxjs_1.Observable.from(this.updater.ensure(uri))
            .map(() => JSON.parse(this.inMemoryFileSystem.getContent(uri))))
            .mergeMap(packageJson => {
            const packageDescriptor = {
                name: packageJson.name,
                version: packageJson.version,
                repoURL: packageJson.repository && packageJson.repository.url || undefined
            };
            // Collect all dependencies for this package.json
            return rxjs_1.Observable.of('dependencies', 'devDependencies', 'peerDependencies', 'optionalDependencies')
                .filter(key => packageJson[key])
                .mergeMap(key => lodash_1.toPairs(packageJson[key]))
                .map(([name, version]) => ({
                attributes: {
                    name,
                    version
                },
                hints: {
                    dependeePackageName: packageJson.name
                }
            }))
                .toArray()
                .map(dependencies => ({
                package: packageDescriptor,
                dependencies
            }));
        })
            .toArray();
    }
    /**
     * Returns all dependencies of a workspace.
     * Superseded by workspace/xpackages
     */
    workspaceXdependencies(params = {}, span = new opentracing_1.Span()) {
        // Ensure package.json files
        return rxjs_1.Observable.from(this.projectManager.ensureModuleStructure())
            .mergeMap(() => this.inMemoryFileSystem.uris())
            .filter(uri => uri.includes('/package.json') && !uri.includes('/node_modules/'))
            .mergeMap(uri => rxjs_1.Observable.from(this.updater.ensure(uri))
            .map(() => JSON.parse(this.inMemoryFileSystem.getContent(uri))))
            .mergeMap(packageJson => rxjs_1.Observable.of('dependencies', 'devDependencies', 'peerDependencies', 'optionalDependencies')
            .filter(key => packageJson[key])
            .mergeMap(key => lodash_1.toPairs(packageJson[key]))
            .map(([name, version]) => ({
            attributes: {
                name,
                version
            },
            hints: {
                dependeePackageName: packageJson.name
            }
        })))
            .toArray();
    }
    /**
     * The Completion request is sent from the client to the server to compute completion items at a
     * given cursor position. Completion items are presented in the
     * [IntelliSense](https://code.visualstudio.com/docs/editor/editingevolved#_intellisense) user
     * interface. If computing full completion items is expensive, servers can additionally provide
     * a handler for the completion item resolve request ('completionItem/resolve'). This request is
     * sent when a completion item is selected in the user interface. A typically use case is for
     * example: the 'textDocument/completion' request doesn't fill in the `documentation` property
     * for returned completion items since it is expensive to compute. When the item is selected in
     * the user interface then a 'completionItem/resolve' request is sent with the selected
     * completion item as a param. The returned completion item should have the documentation
     * property filled in.
     */
    textDocumentCompletion(params, span = new opentracing_1.Span()) {
        return __awaiter(this, void 0, void 0, function* () {
            // Ensure files needed to suggest completions are fetched
            yield this.projectManager.ensureReferencedFiles(params.textDocument.uri, undefined, undefined, span).toPromise();
            const fileName = util.uri2path(params.textDocument.uri);
            const configuration = this.projectManager.getConfiguration(fileName);
            configuration.ensureBasicFiles();
            const sourceFile = this._getSourceFile(configuration, fileName);
            if (!sourceFile) {
                return vscode_languageserver_1.CompletionList.create();
            }
            const offset = ts.getPositionOfLineAndCharacter(sourceFile, params.position.line, params.position.character);
            const completions = configuration.getService().getCompletionsAtPosition(fileName, offset);
            if (completions == null) {
                return vscode_languageserver_1.CompletionList.create();
            }
            return vscode_languageserver_1.CompletionList.create(completions.entries.map(item => {
                const ret = vscode_languageserver_1.CompletionItem.create(item.name);
                const kind = completionKinds[item.kind];
                if (kind) {
                    ret.kind = kind;
                }
                if (item.sortText) {
                    ret.sortText = item.sortText;
                }
                const details = configuration.getService().getCompletionEntryDetails(fileName, offset, item.name);
                if (details) {
                    ret.documentation = ts.displayPartsToString(details.documentation);
                    ret.detail = ts.displayPartsToString(details.displayParts);
                }
                return ret;
            }));
        });
    }
    /**
     * The signature help request is sent from the client to the server to request signature
     * information at a given cursor position.
     */
    textDocumentSignatureHelp(params, span = new opentracing_1.Span()) {
        return __awaiter(this, void 0, void 0, function* () {
            // Ensure files needed to resolve signature are fetched
            yield this.projectManager.ensureReferencedFiles(params.textDocument.uri, undefined, undefined, span).toPromise();
            const filePath = util.uri2path(params.textDocument.uri);
            const configuration = this.projectManager.getConfiguration(filePath);
            configuration.ensureBasicFiles();
            const sourceFile = this._getSourceFile(configuration, filePath);
            if (!sourceFile) {
                throw new Error(`expected source file ${filePath} to exist in configuration`);
            }
            const offset = ts.getPositionOfLineAndCharacter(sourceFile, params.position.line, params.position.character);
            const signatures = configuration.getService().getSignatureHelpItems(filePath, offset);
            if (!signatures) {
                return { signatures: [], activeParameter: 0, activeSignature: 0 };
            }
            const signatureInformations = signatures.items.map(item => {
                const prefix = ts.displayPartsToString(item.prefixDisplayParts);
                const params = item.parameters.map(p => ts.displayPartsToString(p.displayParts)).join(', ');
                const suffix = ts.displayPartsToString(item.suffixDisplayParts);
                const parameters = item.parameters.map(p => {
                    return vscode_languageserver_1.ParameterInformation.create(ts.displayPartsToString(p.displayParts), ts.displayPartsToString(p.documentation));
                });
                return vscode_languageserver_1.SignatureInformation.create(prefix + params + suffix, ts.displayPartsToString(item.documentation), ...parameters);
            });
            return {
                signatures: signatureInformations,
                activeSignature: signatures.selectedItemIndex,
                activeParameter: signatures.argumentIndex
            };
        });
    }
    /*
     * walkMostAST walks most of the AST (the part that matters for gathering all references)
     */
    _walkMostAST(node, visit) {
        visit(node);
        const children = [];
        switch (node.kind) {
            case ts.SyntaxKind.QualifiedName: {
                const n = node;
                children.push(n.left, n.right);
                break;
            }
            case ts.SyntaxKind.ComputedPropertyName: {
                const n = node;
                children.push(n.expression);
                break;
            }
            case ts.SyntaxKind.TypeParameter: {
                const n = node;
                pushall(children, n.name, n.constraint, n.expression);
                break;
            }
            case ts.SyntaxKind.Parameter: {
                const n = node;
                pushall(children, n.name, n.type, n.initializer);
                break;
            }
            case ts.SyntaxKind.Decorator: {
                const n = node;
                children.push(n.expression);
                break;
            }
            case ts.SyntaxKind.PropertySignature: {
                const n = node;
                pushall(children, n.name, n.type, n.initializer);
                break;
            }
            case ts.SyntaxKind.PropertyDeclaration: {
                const n = node;
                pushall(children, n.name, n.type, n.initializer);
                break;
            }
            case ts.SyntaxKind.MethodSignature: {
                const n = node;
                pushall(children, n.name, n.type);
                if (n.typeParameters) {
                    children.push(...n.typeParameters);
                }
                if (n.parameters) {
                    children.push(...n.parameters);
                }
                break;
            }
            case ts.SyntaxKind.MethodDeclaration: {
                const n = node;
                pushall(children, n.name, n.body);
                break;
            }
            case ts.SyntaxKind.Constructor: {
                const n = node;
                pushall(children, n.name, n.body);
                break;
            }
            case ts.SyntaxKind.GetAccessor: {
                const n = node;
                children.push(n.name, n.body);
                break;
            }
            case ts.SyntaxKind.SetAccessor: {
                const n = node;
                children.push(n.name, n.body);
                break;
            }
            case ts.SyntaxKind.CallSignature: {
                const n = node;
                pushall(children, n.name, n.type);
                if (n.typeParameters) {
                    children.push(...n.typeParameters);
                }
                if (n.parameters) {
                    children.push(...n.parameters);
                }
                break;
            }
            case ts.SyntaxKind.ConstructSignature: {
                const n = node;
                pushall(children, n.name, n.type);
                if (n.typeParameters) {
                    children.push(...n.typeParameters);
                }
                if (n.parameters) {
                    children.push(...n.parameters);
                }
                break;
            }
            case ts.SyntaxKind.IndexSignature: {
                const n = node;
                pushall(children, n.name, n.type);
                if (n.typeParameters) {
                    children.push(...n.typeParameters);
                }
                if (n.parameters) {
                    children.push(...n.parameters);
                }
                break;
            }
            case ts.SyntaxKind.TypePredicate: {
                const n = node;
                children.push(n.parameterName, n.type);
                break;
            }
            case ts.SyntaxKind.TypeReference: {
                const n = node;
                children.push(n.typeName);
                if (n.typeArguments) {
                    children.push(...n.typeArguments);
                }
                break;
            }
            case ts.SyntaxKind.ConstructorType:
            case ts.SyntaxKind.FunctionType: {
                const n = node;
                pushall(children, n.name, n.type);
                pushall(children, n.name, n.type);
                if (n.typeParameters) {
                    children.push(...n.typeParameters);
                }
                if (n.parameters) {
                    children.push(...n.parameters);
                }
                break;
            }
            case ts.SyntaxKind.TypeQuery: {
                const n = node;
                children.push(n.exprName);
                break;
            }
            case ts.SyntaxKind.TypeLiteral: {
                const n = node;
                pushall(children, n.name);
                children.push(...n.members);
                break;
            }
            case ts.SyntaxKind.ArrayType: {
                const n = node;
                children.push(n.elementType);
                break;
            }
            case ts.SyntaxKind.TupleType: {
                const n = node;
                children.push(...n.elementTypes);
                break;
            }
            case ts.SyntaxKind.IntersectionType:
            case ts.SyntaxKind.UnionType: {
                const n = node;
                children.push(...n.types);
                break;
            }
            case ts.SyntaxKind.ParenthesizedType: {
                const n = node;
                children.push(n.type);
                break;
            }
            case ts.SyntaxKind.LiteralType: {
                const n = node;
                children.push(n.literal);
                break;
            }
            case ts.SyntaxKind.ObjectBindingPattern:
            case ts.SyntaxKind.ArrayBindingPattern: {
                const n = node;
                children.push(...n.elements);
                break;
            }
            case ts.SyntaxKind.BindingElement: {
                const n = node;
                pushall(children, n.propertyName, n.name, n.initializer);
                break;
            }
            case ts.SyntaxKind.ArrayLiteralExpression: {
                const n = node;
                children.push(...n.elements);
                break;
            }
            case ts.SyntaxKind.ObjectLiteralExpression: {
                const n = node;
                children.push(...n.properties);
                break;
            }
            case ts.SyntaxKind.PropertyAccessExpression: {
                const n = node;
                children.push(n.expression, n.name);
                break;
            }
            case ts.SyntaxKind.ElementAccessExpression: {
                const n = node;
                pushall(children, n.expression, n.argumentExpression);
                break;
            }
            case ts.SyntaxKind.CallExpression: {
                const n = node;
                pushall(children, n.name, n.expression, ...n.arguments);
                if (n.typeArguments) {
                    children.push(...n.typeArguments);
                }
                break;
            }
            case ts.SyntaxKind.NewExpression: {
                const n = node;
                pushall(children, n.name, n.expression, ...n.arguments);
                if (n.typeArguments) {
                    children.push(...n.typeArguments);
                }
                break;
            }
            case ts.SyntaxKind.TaggedTemplateExpression: {
                const n = node;
                children.push(n.tag, n.template);
                break;
            }
            case ts.SyntaxKind.TypeAssertionExpression: {
                const n = node;
                children.push(n.type, n.expression);
                break;
            }
            case ts.SyntaxKind.ParenthesizedExpression: {
                const n = node;
                children.push(n.expression);
                break;
            }
            case ts.SyntaxKind.FunctionExpression: {
                const n = node;
                pushall(children, n.name, n.body);
                break;
            }
            case ts.SyntaxKind.ArrowFunction: {
                const n = node;
                children.push(n.body);
                break;
            }
            case ts.SyntaxKind.DeleteExpression: {
                const n = node;
                children.push(n.expression);
                break;
            }
            case ts.SyntaxKind.TypeOfExpression: {
                const n = node;
                children.push(n.expression);
                break;
            }
            case ts.SyntaxKind.VoidExpression: {
                const n = node;
                children.push(n.expression);
                break;
            }
            case ts.SyntaxKind.AwaitExpression: {
                const n = node;
                children.push(n.expression);
                break;
            }
            case ts.SyntaxKind.PrefixUnaryExpression: {
                const n = node;
                children.push(n.operand);
                break;
            }
            case ts.SyntaxKind.PostfixUnaryExpression: {
                const n = node;
                children.push(n.operand);
                break;
            }
            case ts.SyntaxKind.BinaryExpression: {
                const n = node;
                children.push(n.left, n.right);
                break;
            }
            case ts.SyntaxKind.ConditionalExpression: {
                const n = node;
                children.push(n.condition, n.whenTrue, n.whenFalse);
                break;
            }
            case ts.SyntaxKind.TemplateExpression: {
                const n = node;
                children.push(n.head, ...n.templateSpans);
                break;
            }
            case ts.SyntaxKind.YieldExpression: {
                const n = node;
                pushall(children, n.expression);
                break;
            }
            case ts.SyntaxKind.SpreadElement: {
                const n = node;
                children.push(n.expression);
                break;
            }
            case ts.SyntaxKind.ClassExpression: {
                const n = node;
                pushall(children, n.name, ...n.members);
                if (n.typeParameters) {
                    children.push(...n.typeParameters);
                }
                if (n.heritageClauses) {
                    children.push(...n.heritageClauses);
                }
                break;
            }
            case ts.SyntaxKind.ExpressionWithTypeArguments: {
                const n = node;
                children.push(n.expression);
                if (n.typeArguments) {
                    children.push(...n.typeArguments);
                }
                break;
            }
            case ts.SyntaxKind.AsExpression: {
                const n = node;
                children.push(n.expression, n.type);
                break;
            }
            case ts.SyntaxKind.NonNullExpression: {
                const n = node;
                children.push(n.expression);
                break;
            }
            case ts.SyntaxKind.TemplateSpan: {
                const n = node;
                children.push(n.expression, n.literal);
                break;
            }
            case ts.SyntaxKind.SemicolonClassElement: {
                const n = node;
                if (n.name) {
                    children.push(n.name);
                }
                break;
            }
            case ts.SyntaxKind.Block: {
                const n = node;
                children.push(...n.statements);
                break;
            }
            case ts.SyntaxKind.VariableStatement: {
                const n = node;
                children.push(n.declarationList);
                break;
            }
            case ts.SyntaxKind.ExpressionStatement: {
                const n = node;
                children.push(n.expression);
                break;
            }
            case ts.SyntaxKind.IfStatement: {
                const n = node;
                pushall(children, n.expression, n.thenStatement, n.elseStatement);
                break;
            }
            case ts.SyntaxKind.DoStatement: {
                const n = node;
                children.push(n.expression, n.statement);
                break;
            }
            case ts.SyntaxKind.WhileStatement: {
                const n = node;
                children.push(n.expression, n.statement);
                break;
            }
            case ts.SyntaxKind.ForStatement: {
                const n = node;
                pushall(children, n.initializer, n.condition, n.incrementor, n.statement);
                break;
            }
            case ts.SyntaxKind.ForInStatement: {
                const n = node;
                children.push(n.initializer, n.expression, n.statement);
                break;
            }
            case ts.SyntaxKind.ForOfStatement: {
                const n = node;
                children.push(n.initializer, n.expression, n.statement);
                break;
            }
            case ts.SyntaxKind.ContinueStatement: {
                const n = node;
                if (n.label) {
                    children.push(n.label);
                }
                break;
            }
            case ts.SyntaxKind.BreakStatement: {
                const n = node;
                if (n.label) {
                    children.push(n.label);
                }
                break;
            }
            case ts.SyntaxKind.ReturnStatement: {
                const n = node;
                if (n.expression) {
                    children.push(n.expression);
                }
                break;
            }
            case ts.SyntaxKind.WithStatement: {
                const n = node;
                children.push(n.expression, n.statement);
                break;
            }
            case ts.SyntaxKind.SwitchStatement: {
                const n = node;
                children.push(n.expression, n.caseBlock);
                break;
            }
            case ts.SyntaxKind.LabeledStatement: {
                const n = node;
                children.push(n.label, n.statement);
                break;
            }
            case ts.SyntaxKind.ThrowStatement: {
                const n = node;
                children.push(n.expression);
                break;
            }
            case ts.SyntaxKind.TryStatement: {
                const n = node;
                pushall(children, n.tryBlock, n.catchClause, n.finallyBlock);
                break;
            }
            case ts.SyntaxKind.VariableDeclaration: {
                const n = node;
                pushall(children, n.name, n.type, n.initializer);
                break;
            }
            case ts.SyntaxKind.VariableDeclarationList: {
                const n = node;
                children.push(...n.declarations);
                break;
            }
            case ts.SyntaxKind.FunctionDeclaration: {
                const n = node;
                pushall(children, n.name, n.body, n.type, ...n.parameters);
                if (n.typeParameters) {
                    children.push(...n.typeParameters);
                }
                break;
            }
            case ts.SyntaxKind.ClassDeclaration: {
                const n = node;
                pushall(children, n.name, ...n.members);
                if (n.typeParameters) {
                    children.push(...n.typeParameters);
                }
                if (n.heritageClauses) {
                    children.push(...n.heritageClauses);
                }
                break;
            }
            case ts.SyntaxKind.InterfaceDeclaration: {
                const n = node;
                children.push(n.name, ...n.members);
                if (n.typeParameters) {
                    children.push(...n.typeParameters);
                }
                if (n.heritageClauses) {
                    children.push(...n.heritageClauses);
                }
                break;
            }
            case ts.SyntaxKind.TypeAliasDeclaration: {
                const n = node;
                children.push(n.name, n.type);
                if (n.typeParameters) {
                    children.push(...n.typeParameters);
                }
                break;
            }
            case ts.SyntaxKind.EnumDeclaration: {
                const n = node;
                children.push(n.name, ...n.members);
                break;
            }
            case ts.SyntaxKind.ModuleDeclaration: {
                const n = node;
                pushall(children, n.name, n.body);
                break;
            }
            case ts.SyntaxKind.ModuleBlock: {
                const n = node;
                children.push(...n.statements);
                break;
            }
            case ts.SyntaxKind.CaseBlock: {
                const n = node;
                children.push(...n.clauses);
                break;
            }
            case ts.SyntaxKind.NamespaceExportDeclaration: {
                const n = node;
                children.push(n.name, n.moduleReference);
                break;
            }
            case ts.SyntaxKind.ImportEqualsDeclaration: {
                const n = node;
                children.push(n.name, n.moduleReference);
                break;
            }
            case ts.SyntaxKind.ImportDeclaration: {
                const n = node;
                pushall(children, n.importClause, n.moduleSpecifier);
                break;
            }
            case ts.SyntaxKind.ImportClause: {
                const n = node;
                pushall(children, n.name, n.namedBindings);
                break;
            }
            case ts.SyntaxKind.NamespaceImport: {
                const n = node;
                children.push(n.name);
                break;
            }
            case ts.SyntaxKind.NamedImports: {
                const n = node;
                children.push(...n.elements);
                break;
            }
            case ts.SyntaxKind.ImportSpecifier: {
                const n = node;
                pushall(children, n.propertyName, n.name);
                break;
            }
            case ts.SyntaxKind.ExportAssignment: {
                const n = node;
                pushall(children, n.name, n.expression);
                break;
            }
            case ts.SyntaxKind.ExportDeclaration: {
                const n = node;
                pushall(children, n.exportClause, n.moduleSpecifier, n.name);
                break;
            }
            case ts.SyntaxKind.NamedExports: {
                const n = node;
                children.push(...n.elements);
                break;
            }
            case ts.SyntaxKind.ExportSpecifier: {
                const n = node;
                pushall(children, n.propertyName, n.name);
                break;
            }
            case ts.SyntaxKind.MissingDeclaration: {
                const n = node;
                if (n.name) {
                    children.push(n.name);
                }
                break;
            }
            case ts.SyntaxKind.ExternalModuleReference: {
                const n = node;
                pushall(children, n.expression);
                break;
            }
            case ts.SyntaxKind.JsxElement: {
                const n = node;
                children.push(n.openingElement, n.closingElement, ...n.children);
                break;
            }
            case ts.SyntaxKind.JsxSelfClosingElement: {
                const n = node;
                children.push(n.tagName, ...n.attributes);
                break;
            }
            case ts.SyntaxKind.JsxOpeningElement: {
                const n = node;
                children.push(n.tagName, ...n.attributes);
                break;
            }
            case ts.SyntaxKind.JsxClosingElement: {
                const n = node;
                children.push(n.tagName);
                break;
            }
            case ts.SyntaxKind.JsxAttribute: {
                const n = node;
                pushall(children, n.name, n.initializer);
                break;
            }
            case ts.SyntaxKind.JsxSpreadAttribute: {
                const n = node;
                children.push(n.expression);
                break;
            }
            case ts.SyntaxKind.JsxExpression: {
                const n = node;
                if (n.expression) {
                    children.push(n.expression);
                }
                break;
            }
            case ts.SyntaxKind.CaseClause: {
                const n = node;
                children.push(n.expression, ...n.statements);
                break;
            }
            case ts.SyntaxKind.DefaultClause: {
                const n = node;
                children.push(...n.statements);
                break;
            }
            case ts.SyntaxKind.HeritageClause: {
                const n = node;
                if (n.types) {
                    children.push(...n.types);
                }
                break;
            }
            case ts.SyntaxKind.CatchClause: {
                const n = node;
                children.push(n.variableDeclaration, n.block);
                break;
            }
            case ts.SyntaxKind.PropertyAssignment: {
                const n = node;
                children.push(n.name, n.initializer);
                break;
            }
            case ts.SyntaxKind.ShorthandPropertyAssignment: {
                const n = node;
                pushall(children, n.name, n.objectAssignmentInitializer);
                break;
            }
            case ts.SyntaxKind.EnumMember: {
                const n = node;
                pushall(children, n.name, n.initializer);
                break;
            }
            case ts.SyntaxKind.SourceFile: {
                const n = node;
                children.push(...n.statements);
                break;
            }
            case ts.SyntaxKind.JSDocTypeExpression: {
                const n = node;
                children.push(n.type);
                break;
            }
            case ts.SyntaxKind.JSDocArrayType: {
                const n = node;
                children.push(n.elementType);
                break;
            }
            case ts.SyntaxKind.JSDocUnionType: {
                const n = node;
                children.push(...n.types);
                break;
            }
            case ts.SyntaxKind.JSDocTupleType: {
                const n = node;
                children.push(...n.types);
                break;
            }
            case ts.SyntaxKind.JSDocNullableType: {
                const n = node;
                children.push(n.type);
                break;
            }
            case ts.SyntaxKind.JSDocNonNullableType: {
                const n = node;
                children.push(n.type);
                break;
            }
            case ts.SyntaxKind.JSDocRecordType: {
                const n = node;
                children.push(n.literal);
                break;
            }
            case ts.SyntaxKind.JSDocRecordMember: {
                const n = node;
                pushall(children, n.name, n.type, n.initializer);
                break;
            }
            case ts.SyntaxKind.JSDocTypeReference: {
                const n = node;
                children.push(n.name, ...n.typeArguments);
                break;
            }
            case ts.SyntaxKind.JSDocOptionalType: {
                const n = node;
                children.push(n.type);
                break;
            }
            case ts.SyntaxKind.JSDocFunctionType: {
                const n = node;
                pushall(children, n.name, n.type, ...n.parameters);
                if (n.typeParameters) {
                    children.push(...n.typeParameters);
                }
                break;
            }
            case ts.SyntaxKind.JSDocVariadicType: {
                const n = node;
                children.push(n.type);
                break;
            }
            case ts.SyntaxKind.JSDocConstructorType: {
                const n = node;
                children.push(n.type);
                break;
            }
            case ts.SyntaxKind.JSDocThisType: {
                const n = node;
                children.push(n.type);
                break;
            }
            case ts.SyntaxKind.JSDocComment: {
                const n = node;
                if (n.tags) {
                    children.push(...n.tags);
                }
                break;
            }
            case ts.SyntaxKind.JSDocTag: {
                const n = node;
                children.push(n.tagName);
                break;
            }
            case ts.SyntaxKind.JSDocParameterTag: {
                const n = node;
                pushall(children, n.typeExpression, n.postParameterName, n.parameterName);
                if (n.preParameterName) {
                    children.push(n.preParameterName);
                }
                break;
            }
            case ts.SyntaxKind.JSDocReturnTag: {
                const n = node;
                children.push(n.typeExpression);
                break;
            }
            case ts.SyntaxKind.JSDocTypeTag: {
                const n = node;
                children.push(n.typeExpression);
                break;
            }
            case ts.SyntaxKind.JSDocTemplateTag: {
                const n = node;
                children.push(...n.typeParameters);
                break;
            }
            case ts.SyntaxKind.JSDocTypedefTag: {
                const n = node;
                pushall(children, n.fullName, n.typeExpression, n.jsDocTypeLiteral);
                if (n.name) {
                    children.push(n.name);
                }
                break;
            }
            case ts.SyntaxKind.JSDocPropertyTag: {
                const n = node;
                children.push(n.name, n.typeExpression);
                break;
            }
            case ts.SyntaxKind.JSDocTypeLiteral: {
                const n = node;
                if (n.jsDocPropertyTags) {
                    children.push(...n.jsDocPropertyTags);
                }
                if (n.jsDocTypeTag) {
                    children.push(n.jsDocTypeTag);
                }
                break;
            }
            case ts.SyntaxKind.JSDocLiteralType: {
                const n = node;
                children.push(n.literal);
                break;
            }
            case ts.SyntaxKind.SyntaxList: {
                const n = node;
                children.push(...n._children);
                break;
            }
            default:
                break;
        }
        for (const child of children) {
            if (child) {
                this._walkMostAST(child, visit);
            }
        }
    }
    /**
     * The document open notification is sent from the client to the server to signal newly opened
     * text documents. The document's truth is now managed by the client and the server must not try
     * to read the document's truth using the document's uri.
     */
    textDocumentDidOpen(params) {
        return __awaiter(this, void 0, void 0, function* () {
            const uri = util.uri2reluri(params.textDocument.uri, this.root);
            // Ensure files needed for most operations are fetched
            yield this.projectManager.ensureReferencedFiles(params.textDocument.uri).toPromise();
            this.projectManager.didOpen(util.uri2path(uri), params.textDocument.text);
        });
    }
    /**
     * The document change notification is sent from the client to the server to signal changes to a
     * text document. In 2.0 the shape of the params has changed to include proper version numbers
     * and language ids.
     */
    textDocumentDidChange(params) {
        return __awaiter(this, void 0, void 0, function* () {
            const uri = util.uri2reluri(params.textDocument.uri, this.root);
            let text = null;
            params.contentChanges.forEach(change => {
                if (change.range || change.rangeLength) {
                    throw new Error('incremental updates in textDocument/didChange not supported for file ' + params.textDocument.uri);
                }
                text = change.text;
            });
            if (!text) {
                return;
            }
            this.projectManager.didChange(util.uri2path(uri), text);
        });
    }
    /**
     * The document save notification is sent from the client to the server when the document was
     * saved in the client.
     */
    textDocumentDidSave(params) {
        return __awaiter(this, void 0, void 0, function* () {
            // Ensure files needed to suggest completions are fetched
            yield this.projectManager.ensureReferencedFiles(params.textDocument.uri).toPromise();
            // TODO don't use "relative" URI
            const uri = util.uri2reluri(params.textDocument.uri, this.root);
            this.projectManager.didSave(util.uri2path(uri));
        });
    }
    /**
     * The document close notification is sent from the client to the server when the document got
     * closed in the client. The document's truth now exists where the document's uri points to
     * (e.g. if the document's uri is a file uri the truth now exists on disk).
     */
    textDocumentDidClose(params) {
        return __awaiter(this, void 0, void 0, function* () {
            // Ensure files needed to suggest completions are fetched
            yield this.projectManager.ensureReferencedFiles(params.textDocument.uri).toPromise();
            // TODO don't use "relative" URI
            const uri = util.uri2reluri(params.textDocument.uri, this.root);
            this.projectManager.didClose(util.uri2path(uri));
        });
    }
    /**
     * Fetches (or creates if needed) source file object for a given file name
     * @param configuration project configuration
     * @param fileName file name to fetch source file for or create it
     */
    _getSourceFile(configuration, fileName) {
        const sourceFile = configuration.getProgram().getSourceFile(fileName);
        if (sourceFile) {
            return sourceFile;
        }
        if (!this.projectManager.hasFile(fileName)) {
            return null;
        }
        configuration.getHost().addFile(fileName);
        configuration.syncProgram();
        return configuration.getProgram().getSourceFile(fileName);
    }
    /**
     * Produces async function that converts ReferenceEntry object to Location
     */
    _transformReference(root, program, includeDeclaration, ref) {
        return (callback) => {
            if (!includeDeclaration && ref.isDefinition) {
                return callback();
            }
            const sourceFile = program.getSourceFile(ref.fileName);
            if (!sourceFile) {
                return callback(new Error('source file "' + ref.fileName + '" does not exist'));
            }
            const start = ts.getLineAndCharacterOfPosition(sourceFile, ref.textSpan.start);
            const end = ts.getLineAndCharacterOfPosition(sourceFile, ref.textSpan.start + ref.textSpan.length);
            callback(undefined, vscode_languageserver_1.Location.create(util.path2uri(root, ref.fileName), {
                start,
                end
            }));
        };
    }
    /**
     * transformNavItem transforms a NavigateToItem instance to a SymbolInformation instance
     */
    _transformNavItem(program, item) {
        const sourceFile = program.getSourceFile(item.fileName);
        if (!sourceFile) {
            throw new Error('source file "' + item.fileName + '" does not exist');
        }
        const start = ts.getLineAndCharacterOfPosition(sourceFile, item.textSpan.start);
        const end = ts.getLineAndCharacterOfPosition(sourceFile, item.textSpan.start + item.textSpan.length);
        return vscode_languageserver_1.SymbolInformation.create(item.name, util.convertStringtoSymbolKind(item.kind), vscode_languageserver_1.Range.create(start.line, start.character, end.line, end.character), this._defUri(item.fileName), item.containerName);
    }
    /**
     * Returns an Iterator for all symbols in a given config
     *
     * Note: This method is not traced because it returns an Iterator that may produce values lazily
     *
     * @param config The ProjectConfiguration to search
     * @param query A text or SymbolDescriptor query
     * @param limit An optional limit that is passed to TypeScript
     * @return Iterator that emits SymbolInformations
     */
    _collectWorkspaceSymbols(config, query, limit = Infinity) {
        config.ensureAllFiles();
        if (query) {
            let items;
            if (typeof query === 'string') {
                // Query by text query
                items = config.getService().getNavigateToItems(query, limit, undefined, false);
            }
            else {
                // Query by name
                const packageName = config.getPackageName();
                const packageDescriptor = packageName && { name: packageName } || undefined;
                items = iterare_1.default(config.getService().getNavigateToItems(query.name || '', limit, undefined, false))
                    .filter(item => util.symbolDescriptorMatch(query, {
                    kind: item.kind,
                    name: item.name,
                    containerKind: item.containerKind,
                    containerName: item.containerName,
                    package: packageDescriptor
                }));
            }
            return iterare_1.default(items)
                .map(item => this._transformNavItem(config.getProgram(), item))
                .filter(symbolInformation => util.isLocalUri(symbolInformation.location.uri));
        }
        else {
            // An empty query uses a different algorithm to iterate all files and aggregate the symbols per-file to get all symbols
            // TODO make all implementations use this? It has the advantage of being streamable and cancellable
            return iterare_1.default(this._getNavigationTreeItems(config)).take(limit);
        }
    }
    /**
     * Transforms definition's file name to URI. If definition belongs to TypeScript library,
     * returns git://github.com/Microsoft/TypeScript URL, otherwise returns file:// one
     */
    _defUri(filePath) {
        if (memfs_1.isTypeScriptLibrary(filePath)) {
            return 'git://github.com/Microsoft/TypeScript?v' + ts.version + '#lib/' + path_.basename(filePath);
        }
        return util.path2uri(this.root, filePath);
    }
    /**
     * Fetches up to limit navigation bar items from given project, flattens them
     */
    _getNavigationTreeItems(configuration) {
        return iterare_1.default(configuration.getProgram().getSourceFiles())
            .filter(sourceFile => !memfs_1.isTypeScriptLibrary(sourceFile.fileName))
            .map(sourceFile => {
            try {
                const tree = configuration.getService().getNavigationTree(sourceFile.fileName);
                return this._flattenNavigationTreeItem(tree, null, sourceFile);
            }
            catch (e) {
                this.logger.error('Could not get navigation tree for file', sourceFile.fileName);
                return [];
            }
        })
            .flatten();
    }
    /**
     * Flattens navigation tree by emitting acceptable NavigationTreeItems as SymbolInformations.
     * Some items (source files, modules) may be excluded
     */
    *_flattenNavigationTreeItem(item, parent, sourceFile) {
        const acceptable = TypeScriptService.isAcceptableNavigationTreeItem(item);
        if (acceptable) {
            yield this._transformNavigationTreeItem(item, parent, sourceFile);
        }
        if (item.childItems) {
            for (const childItem of item.childItems) {
                yield* this._flattenNavigationTreeItem(childItem, acceptable ? item : null, sourceFile);
            }
        }
    }
    /**
     * Transforms NavigationTree to SymbolInformation
     */
    _transformNavigationTreeItem(item, parent, sourceFile) {
        const span = item.spans[0];
        const start = ts.getLineAndCharacterOfPosition(sourceFile, span.start);
        const end = ts.getLineAndCharacterOfPosition(sourceFile, span.start + span.length);
        return {
            name: item.text,
            kind: util.convertStringtoSymbolKind(item.kind),
            location: {
                uri: this._defUri(sourceFile.fileName),
                range: { start, end }
            },
            containerName: parent ? parent.text : ''
        };
    }
    /**
     * @return true if navigation tree item is acceptable for inclusion into workspace/symbols
     */
    static isAcceptableNavigationTreeItem(item) {
        // modules and source files should be excluded
        if ([ts.ScriptElementKind.moduleElement, 'sourcefile'].indexOf(item.kind) >= 0) {
            return false;
        }
        // special items may start with ", (, [, or <
        if (/^[<\(\[\"]/.test(item.text)) {
            return false;
        }
        // magic words
        if (['default', 'constructor', 'new()'].indexOf(item.text) >= 0) {
            return false;
        }
        return true;
    }
}
exports.TypeScriptService = TypeScriptService;
function pushall(arr, ...elems) {
    for (const e of elems) {
        if (e) {
            arr.push(e);
        }
    }
    return arr.length;
}
/**
 * Maps string-based CompletionEntry::kind to enum-based CompletionItemKind
 */
const completionKinds = {
    class: vscode_languageserver_1.CompletionItemKind.Class,
    constructor: vscode_languageserver_1.CompletionItemKind.Constructor,
    enum: vscode_languageserver_1.CompletionItemKind.Enum,
    field: vscode_languageserver_1.CompletionItemKind.Field,
    file: vscode_languageserver_1.CompletionItemKind.File,
    function: vscode_languageserver_1.CompletionItemKind.Function,
    interface: vscode_languageserver_1.CompletionItemKind.Interface,
    keyword: vscode_languageserver_1.CompletionItemKind.Keyword,
    method: vscode_languageserver_1.CompletionItemKind.Method,
    module: vscode_languageserver_1.CompletionItemKind.Module,
    property: vscode_languageserver_1.CompletionItemKind.Property,
    reference: vscode_languageserver_1.CompletionItemKind.Reference,
    snippet: vscode_languageserver_1.CompletionItemKind.Snippet,
    text: vscode_languageserver_1.CompletionItemKind.Text,
    unit: vscode_languageserver_1.CompletionItemKind.Unit,
    value: vscode_languageserver_1.CompletionItemKind.Value,
    variable: vscode_languageserver_1.CompletionItemKind.Variable
};
//# sourceMappingURL=typescript-service.js.map